#include "kbd.h"
#include "consola.h"

// Lectura del teclado

static struct {
	unsigned char shift:1, ctrl:1, alt:1;
} kbd_mods;

static union {
	unsigned char ch;
	struct {unsigned char SL:1, NL:1, CL:1, zero:5;} fl;
} kbd_lights;

/*
 * KBDUS means US Keyboard Layout. This is a scancode table
 * used to layout a standard US keyboard.
 */
static unsigned char keymap_us[128] =
{
    0,  27, '1', '2', '3', '4', '5', '6', '7', '8',	/* 9 */
  '9', '0', '-', '=', '\b',	/* Backspace */
  '\t',			/* Tab */
  'q', 'w', 'e', 'r',	/* 19 */
  't', 'y', 'u', 'i', 'o', 'p', '[', ']', '\n',	/* Enter key */
    KEY_CTRL,		/* 29   - Control */
  'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';',	/* 39 */
 '\'', '`',  KEY_SHIFT,	/* Left shift */
 '\\', 'z', 'x', 'c', 'v', 'b', 'n',			/* 49 */
  'm', ',', '.', '/', KEY_SHIFT,	/* Right shift */
  '*',
    KEY_ALT,		/* Alt */
  ' ',	/* Space bar */
    KEY_CAPSLOCK,	/* Caps lock */
    KEY_F1,	/* 59 - F1 key ... > */
    KEY_F2, KEY_F3, KEY_F4, KEY_F5, KEY_F6, KEY_F7, KEY_F8, KEY_F9,
    KEY_F10,	/* < ... F10 */
    KEY_NUMLOCK,	/* 69 - Num lock*/
    KEY_SCRLLOCK,	/* Scroll Lock */
    0,	/* Home key */
    0,	/* Up Arrow */
    KEY_PGUP,	/* Page Up */
  '-',
    0,	/* Left Arrow */
    0,
    0,	/* Right Arrow */
  '+',
    0,	/* 79 - End key*/
    0,	/* Down Arrow */
    KEY_PGDOWN,	/* Page Down */
    0,	/* Insert Key */
    0,	/* Delete Key */
    0,   0,   0,
    KEY_F11,	/* F11 Key */
    KEY_F12,	/* F12 Key */
    0,	/* All other keys are undefined */
};

// Same as before, with SHIFT pressed
static unsigned char keymap_us_shift[128] =
{
    0,  27, '!', '@', '#', '$', '%', '^', '&', '*',	/* 9 */
  '(', ')', '_', '+', '\b',	/* Backspace */
  '\t',			/* Tab */
  'Q', 'W', 'E', 'R',	/* 19 */
  'T', 'Y', 'U', 'I', 'O', 'P', '{', '}', '\n',	/* Enter key */
    KEY_CTRL,		/* 29   - Control */
  'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', ':',	/* 39 */
 '"', '~',  KEY_SHIFT,	/* Left shift */
 '|', 'Z', 'X', 'C', 'V', 'B', 'N',			/* 49 */
  'M', '<', '>', '?', KEY_SHIFT,	/* Right shift */
  '*',
    KEY_ALT,		/* Alt */
  ' ',	/* Space bar */
    KEY_CAPSLOCK,	/* Caps lock */
    KEY_F1,	/* 59 - F1 key ... > */
    KEY_F2, KEY_F3, KEY_F4, KEY_F5, KEY_F6, KEY_F7, KEY_F8, KEY_F9,
    KEY_F10,	/* < ... F10 */
    KEY_NUMLOCK,	/* 69 - Num lock*/
    KEY_SCRLLOCK,	/* Scroll Lock */
    0,	/* Home key */
    0,	/* Up Arrow */
    KEY_PGUP,	/* Page Up */
  '-',
    0,	/* Left Arrow */
    0,
    0,	/* Right Arrow */
  '+',
    0,	/* 79 - End key*/
    0,	/* Down Arrow */
    KEY_PGDOWN,	/* Page Down */
    0,	/* Insert Key */
    0,	/* Delete Key */
    0,   0,   0,
    KEY_F11,	/* F11 Key */
    KEY_F12,	/* F12 Key */
    0,	/* All other keys are undefined */
};

static void kbd_updatelights() {
	// Espero a que se desocupe el controlador
	while((inb(0x64) & 2) != 0)
		;
	outb(0x60, 0xED);
	outb(0x60, kbd_lights.ch);
}

void kbd_readkey() {
	unsigned char scancode, *keymap, key;

	scancode = inb(0x60);

	// Si SHIFT esta presionado, cambiamos el mapa
	keymap = kbd_mods.shift ? keymap_us_shift : keymap_us;

	// Break code (la tecla dejo de ser presionada)
	if (scancode & 0x80) {
		key = keymap[scancode & ~0x80];
		switch (key) {
		case KEY_SHIFT:
			kbd_mods.shift = 0;
			break;
		case KEY_CTRL:
			kbd_mods.ctrl = 0;
			break;
		case KEY_ALT:
			kbd_mods.alt = 0;
			break;
		}
		return;
	} else {
		key = keymap[scancode];
		switch (key) {
		case KEY_SHIFT:
			kbd_mods.shift = 1;
			return;
		case KEY_CTRL:
			kbd_mods.ctrl = 1;
			return;
		case KEY_ALT:
			kbd_mods.alt = 1;
			return;
		case KEY_NUMLOCK:
			kbd_lights.fl.NL ^= 1;
			kbd_updatelights();
			return;
		case KEY_CAPSLOCK:
			kbd_lights.fl.CL ^= 1;
			kbd_updatelights();
			return;
		case KEY_SCRLLOCK:
			kbd_lights.fl.SL ^= 1;
			kbd_updatelights();
			return;
		default:
			// Letras, cambiar segun Caps Lock
			if (kbd_lights.fl.CL) {
				if (key >= 'a' && key <= 'z')
					key += ('A'-'a');
				if (key >= 'A' && key <= 'Z')
					key -= ('A'-'a');
			}
			break;
		}

		// Teclado numerico
		if (!isprint(key) && kbd_lights.fl.NL)
			switch (scancode) {
			case 71:
			case 72:
			case 73: // 7-9
				key = scancode - 16;
				break;
			case 75:
			case 76:
			case 77: // 4-6
				key = scancode - 23;
				break;
			case 79:
			case 80:
			case 81: // 1-3
				key = scancode - 30;
				break;
			case 82:
				key = '0';
				break;
			case 83:
				key = '.';
				break;
			}
	}

	if (key != 0)
		cons_haytecla(key);
}
