#include "loader.h"

#include "mmu.h"
#include "tss.h"
#include "klib.h"
#include "gdt.h"
#include "i386.h"
#include "consola.h"
#include "pak.h"

int num_procesos;
proceso_t procesos[MAX_PROCESOS];

void loader_inicializar() {
	// Vaciar la lista de procesos
	for (int i = 0; i < MAX_PROCESOS; i++)
		SET_INIT(procesos[i].flags);
	num_procesos = 0;

	pak_init();
	int pid;

	pid = loader_exec_batch("t_idle.tsk", 1);
	kassert(pid == PID_IDLE);
	SET_ADD(procesos[pid].flags, SCH_OCULTA);

	pid = loader_exec("t_shell.tsk", 10);
	kassert(pid == PID_SHELL);
}

int loader_exec_batch(const char *name, int quantum) {
	int pid, gdt_ind;
	kassert(num_procesos >= 0 && num_procesos < MAX_PROCESOS);

	int oldcr3 = rcr3();
	lcr3((int)dirPagsKernel);
	
	// Llenar un lugar de la tabla
	for (pid = 0; pid < MAX_PROCESOS; pid++)
		if (!SET_TEST(procesos[pid].flags, SCH_PRESENTE))
			break;
	gdt_ind = pid + GDT_IND_PROCESOS;
	num_procesos++;

	SET_ADD(procesos[pid].flags, SCH_PRESENTE);
	procesos[pid].name = name;
	procesos[pid].quantum = quantum;
	procesos[pid].pila_usuario = obtener_pagina_usuario();
	procesos[pid].pila_kernel = obtener_pagina_kernel();
	procesos[pid].consola = 0;

	// Construir el mapa de memoria de la tarea
	procesos[pid].dir_pags = crear_dir_usuario();
	mmu_map(procesos[pid].dir_pags, procesos[pid].pila_usuario, procesos[pid].pila_usuario, PROT_USER_RW);
	mmu_map(procesos[pid].dir_pags, procesos[pid].pila_kernel, procesos[pid].pila_kernel, PROT_KERNEL_RW);

	// Copio el codigo a donde se va a ejecutar
	int numPags = (pak_find(name)->size + TAMANO_PAGINA) / TAMANO_PAGINA;
	kassert(numPags > 0 && numPags < MAX_PROC_PAGES);
	for (int i = 0; i < numPags; i++) {
		procesos[pid].codigo[i] = obtener_pagina_usuario();
		mmu_map(procesos[pid].dir_pags, (void*)(i*TAMANO_PAGINA), procesos[pid].codigo[i], PROT_USER_RW);
	}
	pak_copyToPages(name, procesos[pid].codigo, numPags);
	procesos[pid].cant_codigo = numPags;
	
	// Entrada en la TSS
	tsss[pid].cs = GDT_SEL_USRCODE;
	tsss[pid].eip = 0;
	tsss[pid].ds = tsss[pid].es = tsss[pid].fs = tsss[pid].gs = tsss[pid].ss = GDT_SEL_USRDATA;
	tsss[pid].ebp = tsss[pid].esp = (unsigned) procesos[pid].pila_usuario + 0xFFF;
	tsss[pid].ss0 = GDT_SEL_KRNDATA;
	tsss[pid].esp0 = (unsigned) procesos[pid].pila_kernel + 0xFFF;
	tsss[pid].cr3 = (unsigned) procesos[pid].dir_pags;
	tsss[pid].eflags = 0x202; //Interrupt enable, y el que va siempre en uno
	//tsss[pid].dtrap = 0;	// XXX: si justo habia un flag "T", sale un "Debug Exception" (numero 1) al hacer task switch
	//tsss[pid].iomap = tsss[pid].ldt = tsss[pid].ptl = 0;

	// Entrada en la GDT
	gdt[gdt_ind].limit_0_15 = 0x67;
	gdt[gdt_ind].base_0_15 =  (unsigned) (&tsss[pid]) & 0xFFFF;
	gdt[gdt_ind].base_23_16 = ((unsigned) (&tsss[pid]) & 0xF0000) >> 16;
	gdt[gdt_ind].type = 9;
	gdt[gdt_ind].s = 0;
	gdt[gdt_ind].dpl = 3;
	//gdt[gdt_ind].dpl = 0;
	gdt[gdt_ind].p = 1;
	gdt[gdt_ind].limit_16_19 = 0;
	gdt[gdt_ind].avl = 0;
	gdt[gdt_ind].l = 0;
	gdt[gdt_ind].db = 1;
	gdt[gdt_ind].g = 0;
	gdt[gdt_ind].base_31_24 = ((unsigned) (&tsss[pid]) & 0xFFF00000) >> 20;

	lcr3(oldcr3);

	return pid;
}

int loader_exec(const char *name, int quantum) {
	int pid = loader_exec_batch(name, quantum);
	procesos[pid].consola = cons_obtenerNueva(pid);
	cons_printf(getcons(), COLOR(COL_MAGENTA, COL_NEGRO),
	    "Tarea \"%s\" con PID %d y tamanio %d ejecutada en la consola %d\n",
	    name, pid, pak_find(name)->size, procesos[pid].consola);

	return pid;
}

void loader_kill(int pid) {
	int gdt_ind = pid + GDT_IND_PROCESOS;

	kassert(pid >= 0 && pid < MAX_PROCESOS && SET_TEST(procesos[pid].flags, SCH_PRESENTE));

	SET_INIT(procesos[pid].flags);
	gdt[gdt_ind].p = 0;
	cons_liberarUsada(procesos[pid].consola);

	for (int i = 0; i < procesos[pid].cant_codigo; i++)
		liberar_pagina_usuario(procesos[pid].codigo[i]);
	liberar_pagina_usuario(procesos[pid].pila_usuario);
	liberar_pagina_kernel(procesos[pid].pila_kernel);
	destruir_dir_usuario(procesos[pid].dir_pags);

	if (pid == PID_SHELL) {
		int tmp = loader_exec("t_shell.tsk", 10);
		kassert(tmp == PID_SHELL);
	}
}

// *** GRAN NOTA ***
// Despues de varias horas, me di cuenta de que parecia colgarse sin razon aparente mas que tardar "demasiado" en actualizar_pantalla()
// sin ejecutar los kpanic() un par de instrucciones mas adelante en kmain(), y se arregla con quantum > 1
int quantum = 10;

/*
 * Devuelve el selector de la GDT de la proxima tarea a correr,
 * o 0 (cero) si seguira ejecutandose la misma.
 */
#define REFRESH_RATE	6
int proxima_tarea(int actual) {
	static int refresh = 0;
	int pid = TR_A_PID(actual), prox = PID_A_TR(PID_IDLE);

	// No hacemos cambio de tarea
	if (quantum > 0 && procesos[pid].flags == SET_NUM2FLAG(SCH_PRESENTE)) {
		quantum--;
		return 0;
	}

	// Buscamos la proxima que se pueda correr
	int nextPid = pid;
	for (int i = 0; i < MAX_PROCESOS - 1; i++) {
		nextPid = (nextPid+1) % MAX_PROCESOS;
		// XXX: No podemos borrarnos solos, tiene que hacerlo otro
		if (SET_TEST(procesos[nextPid].flags, SCH_TERMINADA) && nextPid != pid)
			loader_kill(nextPid);
		else if (procesos[nextPid].flags == SET_NUM2FLAG(SCH_PRESENTE)) {
			quantum = procesos[nextPid].quantum - 1;
			prox = PID_A_TR(nextPid);
			break;
		}
	}

	if (refresh % REFRESH_RATE == 0)
		cons_actualizar();
	refresh++;

	if (prox == actual)
		return 0;
	//printf("Cambiando a PID %d, en refresh %x (flags %b)\n", TR_A_PID(prox), refresh, procesos[TR_A_PID(prox)].flags);
	return prox;
}

int sched_block(int pid) {
	kassert(pid >= 0 && pid < MAX_PROCESOS && SET_TEST(procesos[pid].flags, SCH_PRESENTE));
	kassert(!SET_TEST(procesos[pid].flags, SCH_BLOQUEADA));

	SET_ADD(procesos[pid].flags, SCH_BLOQUEADA);

	// Si me estoy bloqueando a mi mismo, debo esperar al proximo tick para que el scheduler me saque
	if (pid == getpid()) {
		sti();
		while (SET_TEST(procesos[pid].flags, SCH_BLOQUEADA))
			;
		return 1;
	}

	return 0;
}

void sched_unblock(int pid) {
	kassert(pid >= 0 && pid < MAX_PROCESOS && SET_TEST(procesos[pid].flags, SCH_PRESENTE));
	kassert(SET_TEST(procesos[pid].flags, SCH_BLOQUEADA));
	kassert(pid != getpid());

	SET_REMOVE(procesos[pid].flags, SCH_BLOQUEADA);
}

void sched_kill(int pid, int status) {
	kassert(pid >= 0 && pid < MAX_PROCESOS && SET_TEST(procesos[pid].flags, SCH_PRESENTE));
	kassert(pid != PID_IDLE);

	SET_ADD(procesos[pid].flags, SCH_TERMINADA);
	cons_printf(procesos[pid].consola, COLOR(FG_LIGHT(COL_AZUL), COL_NEGRO),
	    "\nTarea %d terminada (exit status %d)\n", pid, status);

	// Si me estoy matando a mi mismo, debo esperar al proximo tick para que el scheduler me saque
	if (pid == getpid()) {
		sti();
		for (;;) ;
	}
}
